/*
 ATtiny85 (DigiSpark) Luefter-Steuerung fuer Kuehlkoerper

 Das BMS (Battery Management System) fuer einen 30zelligen Akku wird heiss, wenn der Balance-Vorgang laeuft.
 Mit einem Radialluefter wird dessen Kuehlblech angeblasen, da der Luefter aber laut ist, wird er nur eingeschaltet,
 wenn er tatschlich benoetigt wird. 
 Die Drehzahl des Luefters wird bei steigender Temperatur ebenfalls erhoeht, falls notwendig bis zur Maximaldrehzahl
 des Luefters.
 Damit der Luefter nicht allzulange mit zu niedriger Spannung gequaelt wird, bevor er anlaeuft, wird ein Offset (EinschaltSchwelle)
  eingebaut, der die PWM erst bei erreichen der notwendigen Anlaufspannung freigibt.
Ein weiterer Grenzwert zum wieder Ausschalten und damit einhergehend eine Hysterese ist ebenfalls implementiert.

 Die Anschluesse am ATTiny:
 ATTiny    Arduino	Info
 Pin  1  - 5		RESET
 Pin  2  - 3		ONEWIREPIN (Anschluss des Temperatursensors DS18B20)
 Pin  3  - 4		externe LED (Option)
 Pin  4  -			GND
 Pin  5  - 0		Luefterdrehzahl (Option)
 Pin  6  - 1		FET driver (PWM) und DigiSpark Onboard-LED
 Pin  7  - 2 / A1	Vsensor (Analog) (Option)
 Pin  8  -			+Vcc

 (Die mit Option gekennzeichneten Funktionen sind nicht implementiert)

 Mit den Standard-Arduino-Einstellungen funktioniert die  Steuerung einwandfrei, der Temperatur-Sensor 
  wird eingelesen, der Lfter wird per PWM passend in der Drehzahl gesteuert.
 Leider pfeift der Lfter unterhalb der vollen Drehzahl jaemmerlich mit der PWM-Frequenz von 4 kHz, das kann so nicht bleiben.
 Im Netz habe ich gefunden, wie man die PWM-Frequenz schneller einstellen kann, aber diese Massnahme aendert den Timer0.
 Der wird auch fuer millis() verwendet, darauf basierende Libraries (z.B. 1wire) funktionieren danach nicht mehr.
 Zeitabhaengige Funktionen mit millis() - dito.
 Der Versuch, den AtTiny mit Spence Kondes "ATTinyCore" dazu zu ueberreden (TIMER_TO_USE_FOR_MILLIS), den Timer1 fuer millis() 
  zu verwenden hat nicht geklappt.
 Es wird zwar tatschlich der Timer1 fuer millis() verwendet, aber das Timing stimmt irgendwie nicht ganz und 1wire laeuft
  ebenfalls nicht (vielleicht mein Fehler, habe ich nicht mehr naeher untersucht).
 Letztlich habe ich die PWM-Erzeugung mittels Timer1 dann von Hand erledigt.
 Wichtig: Die PWM des Timer1 arbeitet mit P1 des AtTiny. Allerdings wird offenbar auch der negierte Ausgang P0 beeinflusst, 
 dieser Pin P0 kann daher nicht fuer den 1wire-Anschluss benutzt werden.
*/
 
#ifdef F_CPU
#undef F_CPU
#endif
#define F_CPU 8000000						// This is used by delay.h library

#include <util/delay.h>						// Adds _delay_ms and _delay_us functions
#include <OneWire.h>

// Defines der Anschluesse
#define DREHZAHL	0						// pruefen, ob der Luefter angelaufen ist (Option)
#define FETdriver	1						// MOSFET Driver output (und OnBoard LED)
#define SPANNUNG	2						// Luefterspannung messen (Option)
#define ONEWIREPIN	3						// Anschluss fuer DS18B20 Temperatursensor
#define LED			4						// LED (Option)
#define RESET		5						// kein nutzbarer IO

// DS18S20 Temperature chip i/o
OneWire TempSensor(ONEWIREPIN);
 
//Timing DS18B20 auslesen
unsigned long oldMillis = 0;						// This holds the previous millisecond count value

// globale Variablen
byte addrR[8];								// Zeichenketten fuer Temperatur auslesen
byte data[12];								// Zeichenketten fuer Temperatur auslesen
int16_t raw;								// Sensorwert
byte cfg;									// Konfiguration des DS18B20 (Aufloesung)
byte i;
byte present = 0;
int ausgabewert = 0;

bool conversionFlag = false;				// laeuft eine Conversion?
unsigned long conversionTimer = 0;			// Zeit nach Start Conversion bis Auslesen der Temperatur
float tempBalancer;							// Kuehlkoerpertemperatur

//#define SpeedUpPWM
#define PWMTOP 255							// 132 fuer 60 kHz, 199 fuer 40 kHz, 255 fuer 31,5 kHz (bei PCK/8)

// Luefter Ein- und Ausschaltverhalten
const unsigned int EinschaltSchwelle = 190;	// erst bei erreichen dieses Wertes wird die PWM aktiviert
const unsigned int AusschaltSchwelle = 160;	// erst bei erreichen dieses Wertes wird die PWM wieder deaktiviert
bool LuefterLaeuft = false;					// Flag fuer laufenden Luefter

void setup()  {  
	// ----- Achtung! Diese Manipulationen an Timer0 crashen die millis()-Funktion ----- 
	// Ohne diese Manipulationen laeuft die PWM mit ~4 kHz, das bringt den Lfter zum Piepsen!!
	// Mit diesen Manipulationen (~32 kHz PWM) an Timer0 ist der Lfter ruhig,  
	//	 allerdings funktioniert dann der 1wire Bus nicht.
#ifdef SpeedUpPWM
	/*
	Control Register A for Timer/Counter-0 (Timer/Counter-0 is configured using two registers: A and B)
	TCCR0A is 8 bits: [COM0A1:COM0A0:COM0B1:COM0B0:unused:unused:WGM01:WGM00]
	2<<COM0A0: sets bits COM0A0 and COM0A1, which (in Fast PWM mode) clears OC0A on compare-match, and sets OC0A at BOTTOM
	2<<COM0B0: sets bits COM0B0 and COM0B1, which (in Fast PWM mode) clears OC0B on compare-match, and sets OC0B at BOTTOM
	3<<WGM00: sets bits WGM00 and WGM01, which (when combined with WGM02 from TCCR0B below) enables Fast PWM mode
	*/
	TCCR0A = 2<<COM0A0 | 2<<COM0B0 | 3<<WGM00;   
	/*
	Control Register B for Timer/Counter-0 (Timer/Counter-0 is configured using two registers: A and B)
	TCCR0B is 8 bits: [FOC0A:FOC0B:unused:unused:WGM02:CS02:CS01:CS00]
	0<<WGM02: bit WGM02 remains clear, which (when combined with WGM00 and WGM01 from TCCR0A above) enables Fast PWM mode
	1<<CS00: sets bits CS01 (leaving CS01 and CS02 clear), which tells Timer/Counter-0 to not use a prescalar
	*/
	TCCR0B = 0<<WGM02 | 1<<CS00;   
	/*
	Control Register for Timer/Counter-1 (Timer/Counter-1 is configured with just one register: this one)
	TCCR1 is 8 bits: [CTC1:PWM1A:COM1A1:COM1A0:CS13:CS12:CS11:CS10]
	0<<PWM1A: bit PWM1A remains clear, which prevents Timer/Counter-1 from using pin OC1A (which is shared with OC0B)
	0<<COM1A0: bits COM1A0 and COM1A1 remain clear, which also prevents Timer/Counter-1 from using pin OC1A (see PWM1A above)
	1<<CS10: sets bit CS11 which tells Timer/Counter-1  to not use a prescalar
	*/
	TCCR1 = 0<<PWM1A | 0<<COM1A0 | 1<<CS10;   
	/*
	General Control Register for Timer/Counter-1 (this is for Timer/Counter-1 and is a poorly named register)
	GTCCR is 8 bits: [TSM:PWM1B:COM1B1:COM1B0:FOC1B:FOC1A:PSR1:PSR0]
	1<<PWM1B: sets bit PWM1B which enables the use of OC1B (since we disabled using OC1A in TCCR1)
	2<<COM1B0: sets bit COM1B1 and leaves COM1B0 clear, which (when in PWM mode) clears OC1B on compare-match, and sets at BOTTOM
	*/
	GTCCR = 1<<PWM1B | 2<<COM1B0;
#endif

	// Alternative: PWM mit Timer1 zu Fuss programmieren, Register initialisieren, PWM auf Pin P1 (FETdriver und OnBoard LED)
	// OCR1C bestimmt zusammen mit dem Prescaler die PWM-Frequenz. OCR1A enthaelt den Vergleichswert fuer Output Compare Match, 
	//  bei Erreichen des Compare-Wertes toggelt Pin P1
	// Register TCCR1: 
	//	CTC1 = 1 (compare match mit OCR1C)
	//	PWM1A = 1 (enable PWM)
	//	COM1A0 = 1	\ toggle P1
	//	COM1A1 = 0	/
	TCCR1 = 0;					// alle Bits loeschen...
	bitSet(TCCR1, CTC1);		// ...dann einzeln setzen soweit notwendig
	bitSet(TCCR1, PWM1A);		//
	bitSet(TCCR1, COM1A0);		//	
	// Prescaler fuer Timer1 und Compare Match Register OCR1C = 255 -> 31,5 kHz PWM
	//	CS10 = 0	\			//
	//	CS11 = 0	 \ PCK/8	//
	//	CS12 = 1	 /			//
	//	CS13 = 0	/			//
	bitSet(TCCR1, CS12);		//
	OCR1C = PWMTOP;				// siehe #define PWMTOP (PWMTOP = 255)

	// PLLCSR - Verwendung des 64 MHz Peripherie-Clocks fuer den PWM-Prescaler
	bitSet(PLLCSR, PLLE);								// PCK enable
	while (bitRead(PLLCSR, PLOCK) != 1);				// warte bis PLL eingerastet ist
	if (bitRead(PLLCSR, PLLE))	bitSet(PLLCSR, PCKE);	// PCKE nur setzen, wenn PLL enabled ist

	// Set up IO pins
	pinMode(SPANNUNG, INPUT);
	pinMode(DREHZAHL, INPUT_PULLUP);
	pinMode(LED, OUTPUT);
	pinMode(FETdriver, OUTPUT);
  
	digitalWrite(FETdriver, LOW);				// FET ausschalten
  
  // Temperatursensor initialisieren
  {
	  // Temperatur
	  if (!TempSensor.search(addrR)) {
		  TempSensor.reset_search();
		  return;
	  }

	  // die nachfolgenden Ausgaben ueber serial.print() sind auf dem DigiSpark nicht realisiert
	  if (OneWire::crc8(addrR, 7) != addrR[7]) {
		  //serial.print("CRC nicht gueltig!\n");
		  return;
	  }

	  if (addrR[0] == 0x10) {
		  //serial.print("Geraet ist aus der DS18S20 Familie.\n");
	  }
	  else if (addrR[0] == 0x28) {
		  //serial.print("Geraet ist aus der DS18B20 Familie.\n");
	  }
	  else {
		  //serial.print("Geraetefamilie nicht erkannt : 0x");
		  //serial.println(addrR[0], HEX);
		  return;
	  }
  }
} 
 
// --------------------------------------------------------------------------------------------------------------
void loop()  { 
 
// die Temperatursensoren DS18b20 benoetigen in der hoechsten Aufloesung ~750 ms fuer die Wandlung.
// Also empfielt es sich, den Start der Wandlung und das Auslesen der Ergebnisse in getrennte
// Routinen zu packen, so dass die Haupt-Loop nicht zyklisch durch warten auf die Ergebnisse 
// ausgebremst wird
	if (conversionFlag == false)	// wenn Wandlung noch nicht laeuft -> starten
	{
		startConversion();
		conversionTimer = millis();					// Startzeit Conversion merken
		conversionFlag = true;
	}

	// wenn Wandlung gestartet wurde und Pause um ist -> Temp auslesen
	if ((conversionFlag == true) && (millis() > (conversionTimer + 800)))
	{
		getTemp();
		conversionFlag = false;						// dadurch kann die naechste Wandlung starten
	}

	// ab hier ist die aktuelle Temperatur mit einer Nachkommastelle bekannt (in tempBalancer aus getTemp())
	ausgabewert = map(tempBalancer * 100.0, 2500.0, 3000.0, 0, PWMTOP);		// Temperatur auf 0..PWMTOP skalieren...
	ausgabewert = constrain(ausgabewert, 0, PWMTOP);						// ...und auf 0..PWMTOP begrenzen

	// nachfolgender Code implemetiert eine Hysteresefunktion
	// bei ca. 36 Grad schaltet der Luefter ein, bei ca. 30 Grad wieder aus, dazwischen wird die Drehzahl gesteuert
	if ((LuefterLaeuft == false) && (ausgabewert > EinschaltSchwelle))		// Luefter einschalten
	{
		OCR1A = ausgabewert;												// den entsprechenden Analogwert ausgeben
		LuefterLaeuft = true;
	}
	else if((LuefterLaeuft == true) && (ausgabewert < AusschaltSchwelle))	// Luefter ausschalten
	{
		OCR1A = 0;															// PWM abschalten
		LuefterLaeuft = false;
	}
	else if((LuefterLaeuft == true))										// Luefter weitelaufen lassen
	{
		OCR1A = ausgabewert;															// PWM abschalten
	}
		
	// Testfunktion fuer Timer0 (millis() Funktion)
	//if (millis() > (oldMillis + 1000))				// Test der millis() Funktion (1 Sekunde Takt)
	//{													
	//	oldMillis = millis();
	//	//digitalWrite(FETdriver, !digitalRead(FETdriver));				// an FETdriver haengt die OnBoard LED
	//	digitalWrite(LED, !digitalRead(LED));							// zusaetzliche LED an P4
	//}
}
// --------------------------------------------------------------------------------------------------------------

// Wandlung der Temperatursensoren starten
void startConversion()
{
	present = TempSensor.reset();
	if (present == 1)							// wenn der Sensor geantwortet hat
	{
		TempSensor.select(addrR);
		TempSensor.write(0x44, 1);				// start Konvertierung, mit power-on am Ende
	}
}

// Temperaturen auslesen
void getTemp()
{
	present = TempSensor.reset();
	if (present == 1)							// wenn der Sensor geantwortet hat
	{
		TempSensor.select(addrR);
		TempSensor.write(0xBE);         		// Wert lesen

		for (i = 0; i < 9; i++) {				// 9 bytes
			data[i] = TempSensor.read();
		}

		// Convert the data to actual temperature
		// because the result is a 16 bit signed integer, it should
		// be stored to an "int16_t" type, which is always 16 bits
		// even when compiled on a 32 bit processor.
		raw = (data[1] << 8) | data[0];

		cfg = (data[4] & 0x60);
		// at lower res, the low bits are undefined, so let's zero them
		if (cfg == 0x00) raw = raw & ~7;		// 9 bit resolution, 93.75 ms
		else if (cfg == 0x20) raw = raw & ~3;	// 10 bit res, 187.5 ms
		else if (cfg == 0x40) raw = raw & ~1;	// 11 bit res, 375 ms
		//	default is 12 bit resolution, 750 ms conversion time

		tempBalancer = (float)raw / 16.0;
	}
}
